# -*- coding:utf-8 -*-
from redis import StrictRedis
import time

__version_info__ = ('0', '2', '0')


class NotAvailable(Exception):
    """ Raised when unable to aquire the Semaphore in non-blocking mode
    """


class Semaphore(object):

    exists_val = 'ok'

    def __init__(self, client, count, namespace=None, stale_client_timeout=None, blocking=True):
        self.client = client or StrictRedis()
        if count < 1:
            raise ValueError("Parameter 'count' must be larger than 1")
        self.count = count
        self.namespace = namespace if namespace else 'SEMAPHORE'
        self.stale_client_timeout = stale_client_timeout
        self.is_use_local_time = False
        self.blocking = blocking
        self._local_tokens = list()

    def _exists_or_init(self):
        old_key = self.client.getset(self.check_exists_key, self.exists_val)
        if old_key:
            return False
        return self._init()

    def _init(self):
        self.client.expire(self.check_exists_key, 10)
        with self.client.pipeline() as pipe:
            pipe.multi()
            pipe.delete(self.grabbed_key, self.available_key)
            pipe.rpush(self.available_key, *range(self.count))
            pipe.execute()
        self.client.persist(self.check_exists_key)

    @property
    def available_count(self):
        return self.client.llen(self.available_key)

    def acquire(self, timeout=0, target=None):
        self._exists_or_init()
        if self.stale_client_timeout is not None:
            self.release_stale_locks()

        if self.blocking:
            pair = self.client.blpop(self.available_key, timeout)
            if pair is None:
                return None
        else:
            pair = self.client.lpop(self.available_key)
            if pair is None:
                raise NotAvailable

        token = pair[1]
        self._local_tokens.append(token)
        self.client.hset(self.grabbed_key, token, self.current_time)
        if target is not None:
            try:
                target(token)
            finally:
                self.signal(token)
        return token

    def release_stale_locks(self, expires=10):
        token = self.client.getset(self.check_release_locks_key, self.exists_val)
        if token:
            return False
        self.client.expire(self.check_release_locks_key, expires)
        try:
            for token, looked_at in self.client.hgetall(self.grabbed_key).iteritems():
                timed_out_at = float(looked_at) + self.stale_client_timeout
                if timed_out_at < self.current_time:
                    self.signal(token)
        finally:
            self.client.delete(self.check_release_locks_key)

    def _is_locked(self, token):
        return self.client.hexists(self.grabbed_key, token)

    def has_lock(self):
        for t in self._local_tokens:
            if self._is_locked(t):
                return True
        return False

    def release(self):
        if not self.has_lock():
            return False
        return self.signal(self._local_tokens.pop())

    def reset(self):
        self._init()

    def signal(self, token):
        if token is None:
            return None
        with self.client.pipeline() as pipe:
            pipe.multi()
            pipe.hdel(self.grabbed_key, token)
            pipe.lpush(self.available_key, token)
            pipe.execute()
            return token

    def get_namespaced_key(self, suffix):
        return '{0}:{1}'.format(self.namespace, suffix)

    @property
    def check_exists_key(self):
        return self._get_and_set_key('_exists_key', 'EXISTS')

    @property
    def available_key(self):
        return self._get_and_set_key('_available_key', 'AVAILABLE')

    @property
    def grabbed_key(self):
        return self._get_and_set_key('_grabbed_key', 'GRABBED')

    @property
    def check_release_locks_key(self):
        return self._get_and_set_key('_release_locks_ley', 'RELEASE_LOCKS')

    def _get_and_set_key(self, key_name, namespace_suffix):
        if not hasattr(self, key_name):
            setattr(self, key_name, self.get_namespaced_key(namespace_suffix))
        return getattr(self, key_name)

    @property
    def current_time(self):
        if self.is_use_local_time:
            return time.time()
        return float('.'.join(map(str, self.client.time())))

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()
        return True if exc_type is None else False



# -*- coding:utf-8 -*-
from __future__ import absolute_import
from unittest import TestCase
from redis import Redis
from redis_semaphore import Semaphore


class SimpleTestCase(TestCase):

    def setUp(self):
        self.client = Redis()
        self.s_count = 2
        self.sem1 = Semaphore(
            client=self.client,
            count=self.s_count
        )
        self.sem1.reset()
        self.sem2 = Semaphore(
            client=self.client,
            count=self.s_count
        )
        self.sem2.reset()

    def test_lock(self):
        assert self.sem1.available_count == self.s_count
        assert self.sem1.acquire() is not None
        assert self.sem1.available_count == (self.s_count - 1)
        self.sem1.release()
        assert self.sem1.available_count == self.s_count

        assert self.sem2.available_count == self.s_count
        assert self.sem1.acquire() is not None
        assert self.sem2.available_count == (self.s_count - 1)
        self.sem1.release()

    def test_with(self):
        assert self.sem1.available_count == self.s_count
        with self.sem1 as sem:
            assert sem.available_count == (self.s_count - 1)
            with sem:
                assert sem.available_count == (self.s_count - 2)
        assert self.sem1.available_count == self.s_count

    def test_create_with_existing(self):
        with self.sem1 as sem1:
            with self.sem2 as sem2:
                assert sem1.available_count == 0
                assert sem2.available_count == 0
                sem3 = Semaphore(client=self.client, count=40)
                assert sem3.available_count == 0

    def test_nonblocking(self):
        from redis_semaphore import NotAvailable
        for _ in range(self.s_count):
            self.sem1.acquire()
        assert self.sem1.available_count == 0

        self.sem1.blocking = False
        with self.assertRaises(NotAvailable):
            with self.sem1:
                assert False, 'should never reach here'


if __name__ == '__main__':
    from os.path import dirname, abspath
    import sys
    import unittest
    d = dirname
    current_path = d(d(abspath(__file__)))
    sys.path.append(current_path)
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os.path import dirname, abspath, join, isfile
import platform

cwd = abspath(dirname(__file__))
module_path = join(cwd, 'redis_semaphore', '__init__.py')
python_version = platform.python_version()
install_requires = []
version_line = [line for line in open(module_path)
                if line.startswith('__version_info__')][0]
__version__ = '.'.join(eval(version_line.split('__version_info__ = ')[-1]))


def install_requirements():
    req_path = join(cwd, 'requirements.txt')
    if isfile(req_path):
        with open(req_path) as f:
            mods = filter(lambda m: len(m) > 0, f.read().split('\n'))
        install_requires.extend(mods)


def read_file(filename):
    path = join(cwd, filename)
    if isfile(path):
        return open(path).read()


install_requirements()


setup(
    name='Redis-Semaphore',
    version=__version__,
    description="A distributed semaphore and mutex built on Redis.",
    long_description=read_file('README.rst'),
    license='MIT',
    author='bluele',
    author_email='jksmphone@gmail.com',
    url='https://github.com/bluele/redis-semaphore',
    classifiers=[
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    ],
    include_package_data=True,
    packages=find_packages(),
    keywords='redis semaphore',
    install_requires=install_requires,
    zip_safe=False
)
